package PureMessageX::Enduser::Auth::Authenticator::Multi;

###############################################################################
# Required inclusions.
###############################################################################
use strict;                         # Activate compile-time syntax checking
use PureMessage::Enduser::Config qw(%config);

###############################################################################
# Derive ourselves from our base class.
###############################################################################
use base qw( PureMessage::Enduser::Auth::Authenticator );

###############################################################################
# Version number.
###############################################################################
our $VERSION = '1.100';

###############################################################################
# Module configuration.
###############################################################################
our %module_config;
our @authenticators;

###############################################################################
# Allow for optional debugging.
###############################################################################
my $DEBUG = $ENV{AUTH_DEBUG};

###############################################################################
# Helper methods to make logging to "httpd_error.log" a bit easier.
###############################################################################
sub log_message
{
    my ($msg) = @_;
    chomp( $msg );
    print STDERR "EU-MULTI-AUTH: $msg\n";
}
sub log_debug
{
    my ($msg) = @_;
    log_message( "[debug] $msg" ) if ($DEBUG);
}
sub log_warning
{
    my ($msg) = @_;
    log_message( "[warn] $msg" );
}
sub log_error
{
    my ($msg) = @_;
    log_message( "[error] $msg" );
}

###############################################################################
# Initialization.
###############################################################################
# Configuration directives:
#   authenticators
#       - required
#       - space-separated list of authenticators to use
#   debug
#       - enables debug logging
###############################################################################
sub init
{
    %module_config = @_;

    ###########################################################################
    # SANITY CHECK: Make sure that our config has all of the required fields.
    ###########################################################################
    foreach my $required (qw( authenticators ))
    {
        if ( (!$module_config{$required}) ||
             ($module_config{$required} =~ /^\s*$/)
           )
        {
            die "EUI Multi Authenticator is missing '$required'.\n" .
                "Multi authentication disabled.\n";
        }
    }

    ###########################################################################
    # Allow for debugging to be specified by the configuration as well as from
    # ENV{AUTH_DEBUG}
    ###########################################################################
    $DEBUG ||= $module_config{'debug'};

    ###########################################################################
    # Initialize all of the authenticators.
    ###########################################################################
    foreach my $authname (@{$module_config{'authenticators'}})
    {
        log_debug( "Initializing authenticator '$authname'..." );

        #######################################################################
        # Get the configuration for this authenticator.
        #######################################################################
        log_debug( "... fetching configuration" );
        my $auth = $config{'Authenticator'}{$authname};
        if (!$auth) {
            die "Authenticator '$authname' unknown.\n";
        }

        #######################################################################
        # Load up the authentication module.
        #######################################################################
        log_debug( "... loading authentication module" );
        eval "require $auth->{module};";
        if ($@) {
            die "Authenticator '$authname' failed to load: $@\n";
        }

        #######################################################################
        # Initialize the authenticator
        #######################################################################
        my $authcfg = $auth->{'config'};
        $authcfg->{'prefix'} = $module_config{'prefix'};
        log_debug( "... initializing" );
        my $fcn;
        eval {
            no strict 'refs';
            $fcn = &{"$auth->{module}::init"}( %{$authcfg} );
        };
        if ($@) {
            die "Authenticator '$authname' failed to initialize; $@\n";
        }

        #######################################################################
        # Save the current (initialized) state of the authenticator, and create
        # a closure that we can use to provide the "multi-ness" that we need.
        #
        # Yup, a little bit of magic here... allows us to not have to worry
        # about the auth modules having package variables that get overwritten
        # when the configuration is loaded.
        #######################################################################
        log_debug( "... saving initialized state" );
        my $state = _save_state( $auth->{module} );

        log_debug( "... creating closure" );
        push( @authenticators,
                [ $authname,
                  sub {
                    my (@args) = @_;
                    _restore_state( $auth->{module}, $state );
                    my %authresults = eval { $fcn->(@args) };
                    log_error( $@ ) if ($@);
                    $state = _save_state( $auth->{module} );
                    return %authresults;
                  },
                ] );
        log_debug( "... done" );
    }
    log_debug( "All authentication handlers initialized" );

    ###########################################################################
    # Return a handle to the authenticator function.
    ###########################################################################
    return \&Authenticate;
}

###############################################################################
# Multi Authentication.
###############################################################################
sub Authenticate
{
    my ($in_user, $in_password, %opts) = @_;

    ###########################################################################
    # Abort if we don't have everything.
    ###########################################################################
    return unless ($in_user and $in_password);

    ###########################################################################
    # Attempt to authenticate the user.
    ###########################################################################
    my %authinfo;
    eval
    {
        #######################################################################
        # Try each of the registered authenticators.
        #######################################################################
        log_debug( "Attempting to authenticate '$in_user'..." );
        foreach my $auth (@authenticators)
        {
            my ($a_name, $a_fcn) = @{$auth};
            log_debug( "... authenticator: $a_name" );

            %authinfo = $a_fcn->( $in_user, $in_password, %opts );
            if (exists $authinfo{'email'}) {
                log_debug( "... ... successful" );
                last;
            } else {
                log_debug( "... ... failed" );
            }
        }

        #######################################################################
        # If we weren't successful, throw an error.
        #######################################################################
        if (!exists $authinfo{'email'}) {
            die "Unable to authenticate user '$in_user'.\n";
        }
    };
    if ($@)
    {
        log_error( $@ );
        return;
    }

    ###########################################################################
    # Authenticated ok; return the auth info back to the caller.
    ###########################################################################
    log_debug( "Authentication successful for '$in_user'." );
    return %authinfo;
}

###############################################################################
# INTERNAL METHOD.  Saves the current state of a packages variables, returning
# them back to the caller in a hash-ref suitable for passing through to
# '_restore_state()'.
#
# Yes, this is a cheap hack, but for our simplistic purposes of needing to be
# able to allow for multiple auth configs, it works.
###############################################################################
sub _save_state
{
    my ($pkg) = @_;
    no strict 'refs', 'vars';
    my $state = {};
    while (my ($var, $val) = each %{"$pkg\:\:"}) {
        *sym = *{"$val"};
        if (defined $sym) {
            $state->{'scalar'}{$var} = $sym;
        } elsif (defined @sym) {
            $state->{'array'}{$var} = [@sym];
        } elsif (defined %sym) {
            $state->{'hash'}{$var} = {%sym};
        }
    }
    return $state;
}

###############################################################################
# INTERNAL METHOD.  Restores the current state of a packages variables, from a
# hash-ref retrieved by calling '_save_state()'.
###############################################################################
sub _restore_state
{
    my ($pkg, $state) = @_;
    no strict 'refs';
    while (my ($var, $val) = each %{$state->{'scalar'}}) {
        ${"$pkg\:\:$var"} = $val;
    }
    while (my ($var, $val) = each %{$state->{'array'}}) {
        @{"$pkg\:\:$var"} = @{$val};
    }
    while (my ($var, $val) = each %{$state->{'hash'}}) {
        %{"$pkg\:\:$var"} = %{$val};
    }
}

###############################################################################
1;
